/* 
   WebDAV XML response handling
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: dav_xml.c,v 1.7 1999/12/30 12:51:10 joe Exp $
*/

#include <config.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <xmlparse.h>

#include "string_utils.h"
#include "http_utils.h"
#include "dav_xml.h"

#define HANDLE_NSPACES

/* TODO: 
 * could move 'valid' into state, maybe allow optional
 * continuation past an invalid branch.
 */

const static struct dav_xml_elm root_element = 
{ "@<root>@", DXE_root, 0 };

/* Linked list of namespace scopes */
struct dav_xml_nspace {
    char *name;
    char *value;
    struct dav_xml_nspace *next;
};

/* The expat handlers */
static void exp_startelm( void *userdata, const char *name, const char **atts );
static void exp_endelm( void *userdata, const char *name );
static void exp_cdata( void *userdata, const char *cdata, int len );

/* And an auxiliary */
static char * parse_element( struct dav_xmlparser *p, 
			     struct dav_xml_state *state,
			     const char *tag, const char **atts );

/* Parses the tag attributes, and handles XML namespaces. 
 * With a little bit of luck.
 * Return zero on success, non-zero on error. */
static char *parse_element( struct dav_xmlparser *p,
			    struct dav_xml_state *state,
			    const char *tag, const char **atts ) {
    struct dav_xml_nspace *ns;
    int attn;
    char *pnt, *ret;
    const char *tag_prefix, *tag_suffix;
    struct dav_xml_state *xmlt;

    DEBUG( DEBUG_XMLPARSE, "Parsing elm of name: [%s]\n", tag );
    /* Parse the atts for namespace declarations */
    for( attn = 0; atts[attn]!=NULL; attn+=2 ) {
	DEBUG( DEBUG_XMLPARSE, "Got attribute: [%s] = [%s]\n", atts[attn], atts[attn+1] );
	if( strcasecmp( atts[attn], "xmlns" ) == 0 ) {
	    /* New default namespace */
	    state->default_ns = strdup( atts[attn+1] );
	    DEBUG( DEBUG_XMLPARSE, "New default namespace: %s\n", 
		   state->default_ns );
	} else if( strncasecmp( atts[attn], "xmlns:", 6 ) == 0 ) {
	    /* New namespace scope */
	    ns = malloc( sizeof( struct dav_xml_nspace ) );
	    ns->next = state->nspaces;
	    state->nspaces = ns;
	    ns->name = strdup( atts[attn]+6 ); /* skip the xmlns= */
	    ns->value = strdup( atts[attn+1] );
	    DEBUG( DEBUG_XMLPARSE, "New namespace scope: %s -> %s\n",
		   ns->name, ns->value );
	}
    }
    /* Now check the tag name for a namespace scope */
    pnt = strchr( tag, ':' );
    tag_prefix = NULL;
    tag_suffix = NULL;
    if( pnt == NULL ) {
	/* No scope - have we got a default? */
	DEBUG( DEBUG_XMLPARSE, "No scope found, searching for default.\n" );
	for( xmlt = state; xmlt!=NULL; xmlt=xmlt->parent ) {
	    if( xmlt->default_ns != NULL ) {
		tag_prefix = xmlt->default_ns;
		break;
	    }
	}
	if( tag_prefix != NULL ) {
	    DEBUG( DEBUG_XMLPARSE, "Found default namespace [%s]\n", tag_prefix );
	} else {
	    DEBUG( DEBUG_XMLPARSE, "No default namespace, using empty.\n" );
	    tag_prefix = "";
	}
	tag_suffix = tag;
    } else {
	DEBUG( DEBUG_XMLPARSE, "Got namespace scope. Trying to resolve..." );
	/* Have a scope - resolve it */
	for( xmlt = state; tag_prefix==NULL && xmlt!=NULL; xmlt=xmlt->parent ) {
	    for( ns = xmlt->nspaces; ns!=NULL; ns=ns->next ) {
		/* Just compare against the bit before the :
		 * pnt points to the colon. */
		if( strncasecmp( ns->name, tag, pnt-tag ) == 0 ) {
		    /* Scope matched! Hoorah */
		    tag_prefix = ns->value;
		    /* end the search */
		    break;
		}
	    }
	}
	if( tag_prefix != NULL ) {
	    DEBUG( DEBUG_XMLPARSE, "Resolved scope to [%s]\n", tag_prefix );
	    /* The suffix is everything after the ':' */
	    tag_suffix = pnt+1;
	    if( *tag_suffix == '\0' ) {
		snprintf( p->error, BUFSIZ, 
			  "Element name missing in '%s' at line %d of response.",
			  tag, XML_GetCurrentLineNumber(p->parser) );
		DEBUG( DEBUG_XMLPARSE, "No element name after ':'. Failed.\n" );
		return NULL;
	    }
	} else {
	    DEBUG( DEBUG_XMLPARSE, "Undeclared namespace.\n" );
	    snprintf( p->error, BUFSIZ, 
		      "Undeclared namespace in '%s' at line %d of response.",
		      tag, XML_GetCurrentLineNumber(p->parser) );
	    return NULL;
	}
    }
    /* here, we have tag_suffix and tag_prefix */
    DEBUG( DEBUG_XMLPARSE, "prefix: [%s], suffix: [%s]\n", tag_prefix, tag_suffix );
    CONCAT2( ret, tag_prefix, tag_suffix );
    DEBUG( DEBUG_XMLPARSE, "You gave me: %s, and I gave you this: %s\n", tag,
	   ret );
    return ret;
}

/* This is the XML_StartElementHandler... called with the start of a
 * new element. */
static void exp_startelm( void *userdata, const char *name, const char **atts ) {
    struct dav_xmlparser *p = (struct dav_xmlparser *)userdata;
    struct dav_xml_state *s;
    char *full_name;
    int n;

    if( !p->valid ) {
	/* We've stopped parsing */
	DEBUG( DEBUG_XML, "Ignoring start of element: %s\n", name );
	return;
    }
    /* Set the new state */
    s = malloc( sizeof(struct dav_xml_state) );
    memset( s, 0, sizeof(struct dav_xml_state) );
    s->parent = p->current;
    p->current = s;
    full_name = parse_element( p, s, name, atts );
    if( !full_name ) {
	/* it bombed. */
	p->valid = 0;
	return;
    }
    /* Map the name to a tag */
    DEBUG( DEBUG_XML, "Mapping tag name... " );
    s->elm = NULL;
    for( n = 0; p->elements[n].name != NULL; n++ ) {
	if( strcasecmp( p->elements[n].name, full_name ) == 0 ) {
	    s->elm = &p->elements[n];
	    break;
	}
    }
    if( s->elm ) {
        DEBUG( DEBUG_XML, "mapped to id %d\n", s->elm->id );
    } else {
	DEBUG( DEBUG_XML, "unknown element.\n" );
	/* FIXME: map to unknown */
    }
    
    /* Do we want cdata? */
    p->want_cdata = ((p->current->elm->flags & DAV_CDATA) == DAV_CDATA);
    
    /* expat is not a validating parser - check the new element
     * is valid in the current context.
     */
    DEBUG( DEBUG_XML, "Checking context of element (parent: %s)\n", 
	   s->parent->elm->name );
    if( (*p->context_cb)( s->parent->elm->id, s->elm->id ) ) {
	DEBUG( DEBUG_XML, "Invalid context.\n" );
	p->valid = 0;
    } else {
	DEBUG( DEBUG_XML, "Valid context.\n" );
    }
    
    free( full_name );

}

/* Destroys given state */
static void destroy_state( struct dav_xml_state *s ) {
    struct dav_xml_nspace *this_ns, *next_ns;
#ifdef HANDLE_NSPACES
    DEBUG( DEBUG_XMLPARSE, "Freeing namespaces...\n" );
    if( s->default_ns!=NULL ) free( s->default_ns );
    /* Free the namespaces */
    this_ns = s->nspaces;
    while( this_ns != NULL ) {
	next_ns = this_ns->next;
	free( this_ns->name );
	free( this_ns->value );
	free( this_ns );
	this_ns = next_ns;
    };
    DEBUG( DEBUG_XMLPARSE, "Finished freeing namespaces.\n" );
#endif
    free( s );
}

static void exp_cdata( void *userdata, const char *data, int len ) {
    struct dav_xmlparser *p = userdata;
    
    if( !p->want_cdata || !p->valid ) return;
    /* First, if this is the beginning of the CDATA, skip all
     * leading whitespace, we don't want it. */
    DEBUG( DEBUG_XML, "Given %d bytes of cdata.\n", len );
    if( p->cdata->used == 0 ) {
	size_t wslen = 0;
	/* Ignore any leading whitespace */
	while( wslen < len && 
	       ( data[wslen] == ' ' || data[wslen] == '\r' ||
		 data[wslen] == '\n' || data[wslen] == '\t' ) ) {
	    wslen++;
	}
	data += wslen;
	len -= wslen;
	DEBUG( DEBUG_XML, "Skipped %d bytes of leading whitespace.\n", 
	       wslen );
	if( len == 0 ) {
	    DEBUG( DEBUG_XML, "Zero bytes of content.\n" );
	    return;
	}
    }
    if( http_buffer_append( p->cdata, data, len ) ) {
	/* OOM */
	p->valid = 0;
    }
}

static void exp_endelm( void *userdata, const char *name ) {
    struct dav_xmlparser *p = userdata;
    struct dav_xml_state *s = p->current;
    if( !p->valid ) {
	/* We've stopped parsing */
	DEBUG( DEBUG_XML, "Parse died. Ignoring end of element: %s\n", name );
	return;
    }
    /* process it */
    if( (*p->element_cb)( p->userdata, s,
			  p->want_cdata?p->cdata->data:NULL ) ) {
	DEBUG( DEBUG_XML, "Element callback failed.\n" );
	p->valid = 0;
    }
    p->current = s->parent;
    /* Move the current pointer up the branch */
    DEBUG( DEBUG_XML, "Back in element: %s\n", p->current->elm->name );
    if( p->want_cdata ) {
	http_buffer_clear( p->cdata );
    } 
    destroy_state( s );
}

int dav_xml_init( struct dav_xmlparser *p, const struct dav_xml_elm *elements, 
		  dav_xml_check_context ctx_cb, 
		  dav_xml_element_cb elm_cb, void *userdata ) {
    /* Initialize the expat stuff */
    memset( p, 0, sizeof( struct dav_xmlparser ) );
    p->parser = XML_ParserCreate( NULL );
    XML_SetElementHandler( p->parser, exp_startelm, exp_endelm );
    XML_SetCharacterDataHandler( p->parser, exp_cdata );
    XML_SetUserData( p->parser, (void *) p );
    /* Initialize other stuff */
    p->valid = 1;
    /* Placeholder for the root element */
    p->current = p->root = malloc( sizeof(struct dav_xml_state) );
    memset( p->root, 0, sizeof(struct dav_xml_state) );
    p->root->elm = &root_element;
    p->elements = elements;
    p->element_cb = elm_cb;
    p->context_cb = ctx_cb;
    p->userdata = userdata;
    /* Initialize the cdata buffer */
    p->cdata = http_buffer_create();
    return 0;
}   

void dav_xml_parse_v( void *userdata, const char *block, size_t len ) {
    struct dav_xmlparser *p = userdata;
    dav_xml_parse( p, block, len );
}

/* Parse the given block of input of length len */
void dav_xml_parse( struct dav_xmlparser *p, const char *block, size_t len ) {
    int ret;
    /* duck out if it's broken */
    if( !p->valid ) {
	DEBUG( DEBUG_XML, "Not parsing %d bytes.\n", len );
	return;
    }
    if( len ) {
	DEBUG( DEBUG_XML, "Parsing %d length buffer.\n", len );
	ret = XML_Parse( p->parser, block, len, 0 );
    } else {
	DEBUG( DEBUG_XML, "Got 0-length buffer, end of response.\n" );
	ret = XML_Parse( p->parser, "", 0, -1 );
    }
    DEBUG( DEBUG_XML, "XML_Parse returned %d\n", ret );
    if( ret == 0 ) {
	snprintf( p->error, BUFSIZ,
		  "XML parse error at line %d of response: %s", 
		  XML_GetCurrentLineNumber(p->parser),
		  XML_ErrorString(XML_GetErrorCode(p->parser)) );
	p->valid = 0;
    }
}

int dav_xml_finish( struct dav_xmlparser *p ) {
    struct dav_xml_state *s, *parent;
    XML_ParserFree( p->parser );
    http_buffer_destroy( p->cdata );
    /* Clean up any states which may remain.
     * If p.valid, then this should be only the root element. */
    for( s = p->current; s!=NULL; s=parent ) {
	parent = s->parent;
	destroy_state( s );
    }
    return !p->valid;
}

