/* 
   WebDAV fetch
   Copyright (C) 1998-99, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: davfetch.c,v 1.2 1999/12/29 14:23:55 joe Exp $
*/

#include <config.h>

#include <sys/types.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include "string_utils.h"
#include "dates.h"
#include "dav_xml.h"
#include "protocol.h"
#include "httpdav.h"
#include "http_utils.h"
#include "uri.h"

/* PROPFIND response parsing */

#define DXE_multistatus 1
#define DXE_response 2
#define DXE_responsedescription 3
#define DXE_href 4
#define DXE_propstat 5
#define DXE_prop 6
#define DXE_status 7
#define DXE_getlastmodified 8
#define DXE_getcontentlength 9
#define DXE_resourcetype 10
#define DXE_collection 11

/* The fetch context is the userdata to the element callback */
struct fetch_context {
    /* Do we want to include the resource identified by the 
     * Request-URI in the fetched list? */
    bool fetch_target;
    /* Temporary store of file info */
    struct proto_file *file;
    /* The complete fetch list */
    struct proto_file *files;
    /* The collection we did a PROPFIND on */
    const char *fetch_root;
};

static int get_depth( const char *href );
static char *strip_href( const char *root, const char *href );
static int
got_element( void *userdata, 
	     const struct dav_xml_state *s, const char *cdata );
static void got_resource( struct fetch_context *ctx );

static void http_get_content_charset( const char *name, const char *value );

char *http_content_charset;

static int check_context( dav_xml_elmid parent, dav_xml_elmid child ) {
    switch( parent ) {
    case DXE_root:
	switch( child ) {
	case DXE_multistatus:
	case DXE_response:
	    return 0;
	default:
	}
	break;
    case DXE_multistatus:
	switch( child ) {
	case DXE_response:
	case DXE_responsedescription:
	    return 0;
	default:
	}
	break;
    case DXE_response:
	switch( child ) {
	case DXE_href: case DXE_propstat: case DXE_responsedescription:
	case DXE_status:
	    return 0;
	default:
	}
	break;
    case DXE_propstat:
	switch( child ) {
	case DXE_prop: case DXE_status: case DXE_responsedescription:
	    return 0;
	default:
	}
	break;
    case DXE_prop:
	switch( child ) {
	case DXE_resourcetype: case DXE_getcontentlength: case DXE_getlastmodified:
	    return 0;
	default:
	}
    case DXE_resourcetype:
	switch( child ) {
	case DXE_collection:
	    return 0;
	default:
	}
    default:
    }
    return -1;
}

/* Returns the depth of the file in the directory heirarchy,
 * i.e. 1 for /foo.html, 2 for /bar/norm.html, 4 for /a/b/c/d.html
 */
static int get_depth( const char *href ) {
    const char *pnt;
    int count = 0;
    for( pnt=href; *pnt != '\0'; pnt++ ) /* oneliner */
	if( *pnt == '/' ) count++;
    return count;
}

static char *strip_href( const char *root, const char *href ) {
    const char *tmp;
    char *dec, *ret;
    size_t rootlen;
    DEBUG( DEBUG_HTTP, "Parsing href [%s]\n", href );
    if( strncmp( href, "http://", 7 ) == 0 ) {
	/* Absolute URI.
	 * Look for the path bit */
	DEBUG( DEBUG_HTTP, "Got absolute URI.\n" );
	tmp = strchr( href+7, '/' );
	if( tmp == NULL ) {
	    DEBUG( DEBUG_HTTP, "No path segment found.\n" );
	    return NULL;
	}
    } else {
	tmp = href;
    }
    DEBUG( DEBUG_HTTP, "Using abspath segment: %s\n", tmp );
    dec = uri_unescape( tmp );
    DEBUG( DEBUG_HTTP, "Decoded is: [%s]. Root is [%s]\n", tmp, root );
    /* Now, dec points to the absPath section of the URI.
     * We check whether this resource is actually in the
     * collection we have done the PROPFIND against.
     */
    rootlen = strlen( root );
    if( strncmp( dec, root, rootlen ) != 0 ) {
	DEBUG( DEBUG_HTTP, "strip_href failed: root collection not matched." );
	ret = NULL;
    } else {
	/* We're in the right place */
	DEBUG( DEBUG_HTTP, "strip_href: Got [%s]\n", dec + rootlen );
	/* Filename must be the basename */
	ret = strdup( dec + rootlen );
    }
    free( dec );
    return ret;
}

static void got_resource( struct fetch_context *ctx ) {
    struct proto_file *current, *previous;
    
#ifdef DEBUGGING
    {
	char *date = rfc1123_date( ctx->file->modtime );
	DEBUG( DEBUG_HTTP, "Got %s resource:\n"
	       "filename = [%s] size = %d mtime = %s\n", 
	       ctx->file->type==proto_dir?"dir":(
		   ctx->file->type==proto_file?"file":"link"),
	       ctx->file->filename, ctx->file->size, date );
	free( date );
    }
#endif DEBUGGING

    if( !ctx->fetch_target && strlen( ctx->file->filename ) == 0 ) {
	DEBUG( DEBUG_HTTP, "Resource is root collection, ignoring.\n" );
	free( ctx->file->filename );
	return;
    }
   
    if( ctx->file->type == proto_dir ) {
	ctx->file->filename[strlen(ctx->file->filename)-1] = '\0';
    }
    DEBUG( DEBUG_HTTP, "Filename is really: %s\n", ctx->file->filename );

    /* Depth in the hierarchy - i.e., how many directories deep the
     * resource is. */
    ctx->file->depth = get_depth( ctx->file->filename );
    DEBUG( DEBUG_HTTP, "Resource is at depth: %d\n", ctx->file->depth );

    /* Insert it into the list, keeping the list sorted by depth */
    for( current=ctx->files, previous=NULL; current!=NULL; 
	 previous=current, current=current->next )
	/* one-liner */
	if( current->depth > ctx->file->depth )
	    break;

    ctx->file->next = current;

    if( previous == NULL ) {
	ctx->files = ctx->file;
    } else {
	previous->next = ctx->file;
    }

    /* Create a new file, ready to be filled in */
    ctx->file = malloc( sizeof( struct proto_file ) );
    memset( ctx->file, 0, sizeof( struct proto_file ) );
}

/* End-of-element handler */
static int
got_element( void *userdata, const struct dav_xml_state *s, const char *cdata ) {
    struct fetch_context *ctx = userdata;
    char *tmp;
    DEBUG( DEBUG_XML, "End of element %s.\n", s->elm->name );
    switch( s->elm->id ) {
    case DXE_href:
	tmp = strip_href( ctx->fetch_root, cdata );
	if( tmp ) {
	    ctx->file->filename = tmp;
	} else {
	    return -1;
	}
	break;
    case DXE_getlastmodified:
	DEBUG( DEBUG_HTTP, "Parsing date [%s]\n", cdata );
	ctx->file->modtime = rfc1123_parse( cdata );
	if( ctx->file->modtime == (time_t)-1 ) {
	    DEBUG( DEBUG_HTTP, "Date is not in RFC1123 format.\n" );
	    return -1;
	}
	break;
    case DXE_getcontentlength:
	ctx->file->size = atoi( cdata );
	break;
    case DXE_collection:
	ctx->file->type = proto_dir;
	break;
    case DXE_response:
	got_resource( ctx );
	break;
    default:
    }
    return 0;
}

static void http_get_content_charset( const char *name, const char *value ) {
    char **pairs;
    int n;
    if( strcasecmp( name, "Content-Type" ) == 0 ) {
	/* Get the charset so we can pass it on to expat */
	pairs = pair_string( value, ';', '=', http_quotes, http_whitespace );
	for( n = 0; pairs[n] != NULL; n+=2 ) {
	    if( (strcasecmp( pairs[n], "charset") == 0) &&
		pairs[n+1] != NULL ) {
		HTTP_FREE( http_content_charset );
		/* Strip off the quotes */
		http_content_charset = shave_string( pairs[n+1], '\"' );
		DEBUG( DEBUG_HTTP, "Got content type charset: %s\n",
		       http_content_charset );
	    }
	}
	pair_string_free( pairs );
    }
    return;
}


/* WebDAV fetch mode handler. */
int dav_fetch( const char *dirname, int depth, int fetch_target,
	       struct proto_file **files ) {
    http_req_t req;
    struct fetch_context ctx;
    struct dav_xmlparser p;
    int ret, pret;
    static const struct dav_xml_elm elms[] = {
	{ "DAV:multistatus", DXE_multistatus, 0 },
	{ "DAV:response", DXE_response, 0 },
	{ "DAV:responsedescription", DXE_responsedescription, 0 },
	{ "DAV:href", DXE_href, DAV_CDATA },
	{ "DAV:propstat", DXE_propstat, 0 },
	{ "DAV:prop", DXE_prop, 0 },
	{ "DAV:status", DXE_prop, DAV_CDATA },
	{ "DAV:getlastmodified", DXE_getlastmodified, DAV_CDATA },
	{ "DAV:getcontentlength", DXE_getcontentlength, DAV_CDATA },
	{ "DAV:resourcetype", DXE_resourcetype, DAV_CDATA },
	{ "DAV:collection", DXE_collection, 0 },
	{ NULL, 0, 0 }
    };

    const char *propfind_body =
	"<?xml version=\"1.0\"?>" EOL /* should we use encoding=? */
	"<propfind xmlns=\"DAV:\">" EOL
	"  <prop>" EOL
	"    <getcontentlength/>" EOL
	"    <getlastmodified/>" EOL
	"    <resourcetype/>" EOL
	"  </prop>" EOL
	"</propfind>" EOL;
    const char *myheaders =
	"Content-Type: text/xml" EOL; /* should we use charset=? */
    
    ctx.fetch_root = dirname;
    ctx.fetch_target = fetch_target;
    ctx.file = malloc( sizeof( struct proto_file ) );
    memset( ctx.file, 0, sizeof( struct proto_file ) );
    ctx.files = NULL;

    if( dav_xml_init( &p, elms, check_context, got_element, (void *) &ctx ) ) {
	return PROTO_ERROR;
    }

    http_request_init( &req, "PROPFIND", dirname );
    req.body_callback = dav_xml_parse_v;
    req.body_callback_userdata = &p;
    req.body = http_body_buffer;
    req.body_buffer = propfind_body;
    /* Add in the content type header */
    strcat( req.headers, myheaders );
    switch( depth ) {
    case 0:
	strcat( req.headers, "Depth: 0" EOL );
	break;
    case 1:
	strcat( req.headers, "Depth: 1" EOL );
	break;
    default:
	strcat( req.headers, "Depth: infinity" EOL );
    }

    ret = http_request( &req );
    pret = dav_xml_finish( &p );
    free( ctx.file );

    if( ret == PROTO_OK && req.class == 2 && !pret ) {
	if( files ) *files = ctx.files;
	ret = PROTO_OK;
    } else {
	if( files ) *files = NULL;
	ret = PROTO_ERROR;
    }
    http_request_end( &req );
    return ret;
}

