/* 
   WebDAV lock handling
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: davlocks.c,v 1.7 1999/12/30 13:26:45 joe Exp $
*/

/* TODO: building XML structure analysis thingys like this is frigging
 * boring, and could be automated.  A tool should read in a DTD-like
 * description and automatically produce the C code to handle the
 * expat callbacks. Maybe I should just give up with this crap and 
 * build complete trees and fiddle with them afterwards.
 */

#include <config.h>

#define _GNU_SOURCE
#include <stdio.h>

#ifndef HAVE_SNPRINTF
#include "snprintf.h"
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <ctype.h>

#include "httpdav.h"
#include "davlocks.h"
#include "dav_xml.h"
#include "http_utils.h"
#include "uri.h"

/* TODO: It's really not this simple.
 * Problems with this approach:
 *  - If we want to do LOCK /foo/ and there is already a lock at /foo/bar,
 *    we need to supply the locktoken for /foo/bar.
 */
char *dav_lock_ifheader( struct dav_lock *list, const char *uri ) {
    struct http_buffer *buf;
    char *ret = NULL;
    struct dav_lock *lock;
    int found = 0;

    if( (buf = http_buffer_create()) == NULL ) return NULL;
    if( http_buffer_zappend( buf, "If:" ) ) {
	free( buf );
	return NULL;
    }
    /* Iterate through the list, building up a tagged-list in buf */
    for( lock = dav_lock_find(list, uri ); 
	 lock != NULL; lock = dav_lock_find( lock->next, uri ) ) {
	DEBUG( DEBUG_LOCKS, "Got: %s on %s\n", lock->token, lock->uri );
#if 1
	if( http_buffer_concat( buf, " <", lock->uri, "> (<", lock->token, ">)", NULL ) )
#else
	if( http_buffer_concat( buf, " (<", lock->token, ">)", NULL ) )
#endif
	    break;
	DEBUG( DEBUG_LOCKS, "Header now: [%s]", buf->data );
	found = 1;
    }
    if( found ) {
	/* FIXME: breaks the abstraction layer */
	http_buffer_zappend( buf, EOL );
	ret = buf->data;
	free( buf );
    } else {
	http_buffer_destroy( buf );
    }
    return ret;
}

/* Finds a lock on URI, including any children if children is non-zero */
struct dav_lock *
dav_lock_find( struct dav_lock *list, const char *uri /* , int children */ ) {
    struct dav_lock *cur;
    DEBUG( DEBUG_LOCKS, "Finding %s\n", uri );
    for( cur = list; cur != NULL; cur = cur->next ) {
	if( cur->depth == 0 ) {
	    DEBUG( DEBUG_LOCKS, "Exact compare with %s\n", cur->uri );
	    if( uri_compare( uri, cur->uri ) == 0 ) return cur;
	} else {
	    DEBUG( DEBUG_LOCKS, "Child compare with %s\n", cur->uri );
	    if( !uri_childof( cur->uri, uri ) ) return cur;
	}
#if 0
	/* need this for MOVE destination, but, this code is all dead.
	 * wait for the 0.10 */
	if( children && !uri_childof( uri, cur->uri ) ) return cur;
#endif
    }
    return NULL;
}

void dav_lock_add( struct dav_lock **list, struct dav_lock *lock ) {
    if( *list ) {
	(*list)->prev = lock;
    }
    lock->prev = NULL;
    lock->next = *list;
    *list = lock;
}

void dav_lock_remove( struct dav_lock **list, struct dav_lock *lock ) {
    if( lock->prev ) {
	lock->prev->next = lock->next;
    } else {
	*list = lock->next;
    }
    if( lock->next ) {
	lock->next->prev = lock->prev;
    }
}

int dav_unlock( struct dav_lock *lock ) {
    http_req_t req;
    int ret;

    http_request_init( &req, "UNLOCK", lock->uri );
    
    strcat( req.headers, "Lock-Token: <" );
    strcat( req.headers, lock->token );
    strcat( req.headers, ">" EOL );
    
    ret = http_request( &req );
    
    if( ret == PROTO_OK && req.class == 2 ) {
	ret = PROTO_OK;    
    } else {
	ret = PROTO_ERROR;
    }
    
    return ret;
}


#define DXE_prop 1
#define DXE_lockdiscovery 2
#define DXE_activelock 3
#define DXE_lockscope 4
#define DXE_locktype 5
#define DXE_depth 6
#define DXE_owner 7
#define DXE_timeout 9
#define DXE_locktoken 10
#define DXE_lockinfo 11
/* oops */
#define DXE_write 14
#define DXE_exclusive 15
#define DXE_shared 16
#define DXE_href 17
#define DXE_multistatus 18
#define DXE_response 19
#define DXE_responsedescription 20
#define DXE_propstat 21
#define DXE_status 22

static const struct dav_xml_elm lock_elms[] = {
#define A(x) { "DAV:" #x, DXE_ ## x, 0 /* DAV_COLLECT */ } /* ANY */
#define D(x) { "DAV:" #x, DXE_ ## x, 0 }                   /* normal */
#define C(x) { "DAV:" #x, DXE_ ## x, DAV_CDATA }           /* (#PCDATA) */
#define E(x) { "DAV:" #x, DXE_ ## x, 0 /* DAV_LEAF */ }    /* EMPTY */
    D(prop), D(lockdiscovery),
    D(activelock),
    D(lockscope), D(locktype), C(depth), A(owner), C(timeout), D(locktoken),
    /* no lockentry */
    D(lockinfo), D(lockscope), D(locktype),
    E(write), E(exclusive), E(shared), C(href), 
    D(multistatus), D(response), D(responsedescription),
    D(propstat),
    C(status),
    { NULL, 0, 0 }
};

struct lock_context {
    struct lock_result *list;
    struct lock_result *current;
};

static int check_context( dav_xml_elmid parent, dav_xml_elmid child) {
    /* TODO: validate. */
    return 0;
}

static int parse_depth( const char *depth ) {
    if( strcasecmp( depth, "infinity" ) == 0 ) {
	return DAV_DEPTH_INFINITE;
    } else if( isdigit( depth[0] ) ) {
	return atoi( depth );
    } else {
	return -1;
    }
}

static int parse_timeout( const char *depth ) {
    if( strcasecmp( depth, "infinite" ) == 0 ) {
	return DAV_TIMEOUT_INFINITE;
    } else if( strncasecmp( depth, "Second-", 7 ) == 0 ) {
        return atoi( depth+7 );
    } else {
	return DAV_TIMEOUT_INVALID;
    }
}

void got_lockresult( struct lock_context *ctx ) {
    ctx->current->next = ctx->list;
    ctx->list = ctx->current;
    ctx->current = malloc( sizeof(struct lock_result) );
    memset( ctx->current, 0, sizeof(struct lock_result) );
}

static int 
got_element( void *userdata, const struct dav_xml_state *s, 
	     const char *cdata ) {
    struct lock_context *ctx = userdata;
    struct lock_result *l = ctx->current;
    switch( s->elm->id ){ 
    case DXE_write:
	l->lock.type = dav_locktype_write;
	break;
    case DXE_exclusive:
	l->lock.scope = dav_lockscope_exclusive;
	break;
    case DXE_shared:
	l->lock.scope = dav_lockscope_shared;
	break;
    case DXE_depth:
	DEBUG( DEBUG_XML, "Got depth: %s\n", cdata );
	l->lock.depth = parse_depth( cdata );
	if( l->lock.depth == -1 ) {
	    return -1;
	}
	break;
    case DXE_timeout:
	l->lock.timeout = parse_timeout( cdata );
	if( l->lock.timeout == 0 ) {
	    return -1;
	}
	break;
    case DXE_href:
	switch( s->parent->elm->id ) {
	case DXE_locktoken:
	    l->lock.token = strdup( cdata );
	    break;
	case DXE_response:
	    l->href = strdup( cdata );
	    break;
	default:
	}
	break;
    case DXE_status:
	l->reason = strdup( cdata );
	break;
    case DXE_activelock:
	got_lockresult( ctx );
	break;
    }
    return 0;
}

int dav_lock( struct dav_lock *lock ) {
    http_req_t req;
    struct http_buffer *body;
    struct dav_xmlparser p;
    struct lock_context ctx = {0};
    int ret;

    if( dav_xml_init( &p, lock_elms, check_context, got_element, (void *)&ctx ) )
	return PROTO_ERROR;
    
    ctx.current = malloc( sizeof(struct lock_result) );
    if( !ctx.current ) 
	return PROTO_ERROR;

    body = http_buffer_create();
    if( !body ) return PROTO_ERROR;

    /* Create the body */
    if( http_buffer_concat( body,
			    "<?xml version=\"1.0\" encoding=\"utf-8\"?>" EOL
			    "<lockinfo xmlns='DAV:'>" EOL " <lockscope>",
			    lock->scope==dav_lockscope_exclusive?
			    "<exclusive/>":"<shared/>",
			    "</lockscope>" EOL
			    "<locktype><write/></locktype>", NULL ) )
	return PROTO_ERROR;
    if( lock->owner ) {
	if( http_buffer_concat( body, "<owner>", lock->owner, 
				"</owner>" EOL, NULL ) )
	    return PROTO_ERROR;
    }
    if( http_buffer_zappend( body, "</lockinfo>" EOL ) )
	return PROTO_ERROR;

    http_request_init( &req, "LOCK", lock->uri );
    req.body = http_body_buffer;
    req.body_buffer = body->data;
    req.body_callback = dav_xml_parse_v;
    req.body_callback_userdata = &p;

    switch( lock->depth ) {
    case 0:
	strcat( req.headers, "Depth: 0" EOL );
	break;
    default:
	strcat( req.headers, "Depth: Infinity" EOL );
	break;
    }

    ret = http_request( &req );

    http_buffer_destroy( body );
    free( ctx.current );

    if( ret == PROTO_OK && req.class == 2 && ctx.list != NULL ) {
	lock->token = ctx.list->lock.token;
	ret = PROTO_OK;
    } else {
	ret = PROTO_ERROR;
    }

    /* free the list */
    return ret;
}

/* Discover all locks on URI */
int dav_lock_discover( const char *uri, struct lock_result **locks ) {
    http_req_t req;
    struct dav_xmlparser p;
    struct lock_context ctx = {0};
    static const char *body = 
	"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
	"<propfind xmlns='DAV:'><prop><lockdiscovery/></prop></propfind>" EOL;
    int ret;

    if( dav_xml_init( &p, lock_elms, check_context, got_element, (void *)&ctx ) )
	return PROTO_ERROR;
    
    ctx.current = malloc( sizeof(struct lock_result) );
    if( !ctx.current ) 
	return PROTO_ERROR;
    memset( ctx.current, 0, sizeof(struct lock_result) );

    http_request_init( &req, "PROPFIND", uri );
    req.body = http_body_buffer;
    req.body_buffer = body;
    req.body_callback = dav_xml_parse_v;
    req.body_callback_userdata = &p;

    strcat( req.headers, "Depth: 0" EOL );

    ret = http_request( &req );

    free( ctx.current );

    if( ret == PROTO_OK && req.class == 2 ) {
	if( !p.valid ) {
	    strcpy( http_error, p.error );
	    ret = PROTO_ERROR;
	} else if( ctx.list != NULL ) {
	    *locks = ctx.list;
	    ret = PROTO_OK;
	} else {
	    ret = PROTO_NONE;
	}
    } else {
	ret = PROTO_ERROR;
    }

    if( ret == PROTO_ERROR ) {
	/* free the list */
    }
    
    dav_xml_finish( &p );

    return ret;
}

